# -*- coding: utf-8 -*-
"""
/***************************************************************************
 WeatherCitizen
                                 A QGIS plugin
 This plugin loads data from the Weather Citizen server
 Generated by Plugin Builder: http://g-sherman.github.io/Qgis-Plugin-Builder/
                              -------------------
        begin                : 2018-04-04
        git sha              : $Format:%H$
        copyright            : (C) 2018 by Rachel Gilmore
        email                : RGilmore@creare.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
from PyQt5.QtCore import QSettings, QTranslator, qVersion, QCoreApplication
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QAction

# Initialize Qt resources from file resources.py
from .resources import *
# Import the code for the dialog
from .weather_citizen_dialog import WeatherCitizenDialog
import os.path
# import libraries specific to this plugin
from qgis.core import QgsRectangle, QgsMapLayer
import math
import requests
import json
from datetime import datetime, timedelta
import pytz

# global variable for collections list:
collections = ["geosensors","buoys"] #, "openweather"]
# global variable for uuids
uuids = ['All']
# global variable for radius of the earth
RADIUS = 6378137.0

class WeatherCitizen:
    """QGIS Plugin Implementation."""

    def __init__(self, iface):
        """Constructor.

        :param iface: An interface instance that will be passed to this class
            which provides the hook by which you can manipulate the QGIS
            application at run time.
        :type iface: QgsInterface
        """
        # Save reference to the QGIS interface
        self.iface = iface
        # initialize plugin directory
        self.plugin_dir = os.path.dirname(__file__)
        # initialize locale
        locale = QSettings().value('locale/userLocale')[0:2]
        locale_path = os.path.join(
            self.plugin_dir,
            'i18n',
            'WeatherCitizen_{}.qm'.format(locale))

        if os.path.exists(locale_path):
            self.translator = QTranslator()
            self.translator.load(locale_path)

            if qVersion() > '4.3.3':
                QCoreApplication.installTranslator(self.translator)

        # Create the dialog (after translation) and keep reference
        self.dlg = WeatherCitizenDialog()

        # Declare instance attributes
        self.actions = []
        self.menu = self.tr(u'&Weather Citizen')
        # TODO: We are going to let the user set this up in a future iteration
        self.toolbar = self.iface.addToolBar(u'WeatherCitizen')
        self.toolbar.setObjectName(u'WeatherCitizen')

    # noinspection PyMethodMayBeStatic
    def tr(self, message):
        """Get the translation for a string using Qt translation API.

        We implement this ourselves since we do not inherit QObject.

        :param message: String for translation.
        :type message: str, QString

        :returns: Translated version of message.
        :rtype: QString
        """
        # noinspection PyTypeChecker,PyArgumentList,PyCallByClass
        return QCoreApplication.translate('WeatherCitizen', message)


    def add_action(
        self,
        icon_path,
        text,
        callback,
        enabled_flag=True,
        add_to_menu=True,
        add_to_toolbar=True,
        status_tip=None,
        whats_this=None,
        parent=None):
        """Add a toolbar icon to the toolbar.

        :param icon_path: Path to the icon for this action. Can be a resource
            path (e.g. ':/plugins/foo/bar.png') or a normal file system path.
        :type icon_path: str

        :param text: Text that should be shown in menu items for this action.
        :type text: str

        :param callback: Function to be called when the action is triggered.
        :type callback: function

        :param enabled_flag: A flag indicating if the action should be enabled
            by default. Defaults to True.
        :type enabled_flag: bool

        :param add_to_menu: Flag indicating whether the action should also
            be added to the menu. Defaults to True.
        :type add_to_menu: bool

        :param add_to_toolbar: Flag indicating whether the action should also
            be added to the toolbar. Defaults to True.
        :type add_to_toolbar: bool

        :param status_tip: Optional text to show in a popup when mouse pointer
            hovers over the action.
        :type status_tip: str

        :param parent: Parent widget for the new action. Defaults None.
        :type parent: QWidget

        :param whats_this: Optional text to show in the status bar when the
            mouse pointer hovers over the action.

        :returns: The action that was created. Note that the action is also
            added to self.actions list.
        :rtype: QAction
        """

        icon = QIcon(icon_path)
        action = QAction(icon, text, parent)
        action.triggered.connect(callback)
        action.setEnabled(enabled_flag)

        if status_tip is not None:
            action.setStatusTip(status_tip)

        if whats_this is not None:
            action.setWhatsThis(whats_this)

        if add_to_toolbar:
            self.toolbar.addAction(action)

        if add_to_menu:
            self.iface.addPluginToWebMenu(
                self.menu,
                action)

        self.actions.append(action)

        return action

    def initGui(self):
        """Create the menu entries and toolbar icons inside the QGIS GUI."""

        icon_path = ':/plugins/weather_citizen/icon.png'
        self.add_action(
            icon_path,
            text=self.tr(u'Load Weather Citizen Data'),
            callback=self.run,
            parent=self.iface.mainWindow())
        # will be set False in run()
        self.first_start = True


    def unload(self):
        """Removes the plugin menu item and icon from QGIS GUI."""
        for action in self.actions:
            self.iface.removePluginWebMenu(
                self.tr(u'&Weather Citizen'),
                action)
            self.iface.removeToolBarIcon(action)
        # remove the toolbar
        del self.toolbar


    def load_map(self):
        # load map if not already loaded upon launching plugin
        found = 0
        for layer in self.iface.mapCanvas().layers():
            if 'Google' in layer.name():
                found = 1
                break
        if found == 0:
            webmenu = self.iface.webMenu()
            cont = 0
            for act in webmenu.actions():
                if 'QuickMapServices' in act.text():
                    qmsmenu = act
                    cont = 1
                    break
            if cont == 1:
                cont = 0
                for act in qmsmenu.menu().actions():
                    if 'Google' in act.text():
                        googlemenu = act
                        cont = 1
                        break
            if cont == 1:
                cont = 0
                for act in googlemenu.menu().actions():
                    if 'Hybrid' in act.text():
                        act.trigger()
                        cont = 1
                        break
            # zoom to default start area (currently NH/VT area)
            if cont == 1:
                canvas = self.iface.mapCanvas()
                canvas.renderComplete.connect(self.zoom_start)


    def zoom_start(self):
        # zoom to NH/VT area upon launching OpenStreetMap
        creare = [-72.233305,43.684772]
        pos = [math.radians(creare[0])*RADIUS, math.log(math.tan(math.pi/4 + math.radians(creare[1])/2))*RADIUS]
        offset = 200000
        zoomRectangle = QgsRectangle(pos[0]-offset, pos[1]-offset,pos[0]+offset,pos[1]+offset)
        canvas = self.iface.mapCanvas()
        canvas.setExtent(zoomRectangle)
        canvas.refresh()
        canvas.renderComplete.disconnect(self.zoom_start)


    def toggle_uuid_display(self):
        # when changing selected data collection, update gui options
        selected_collection_idx = self.dlg.collectionBox.currentIndex()
        selected_collection = collections[selected_collection_idx]
        # show device uuid info if collection is "geosensors", hide otherwise
        if selected_collection == "geosensors":
            self.dlg.uuid_label.setEnabled(True)
            self.dlg.uuid_manual_label.setEnabled(True)
            self.dlg.uuid_manual_entry.setEnabled(True)
            self.dlg.uuid_horizontalSlider.setEnabled(True)
            self.dlg.uuid_dropdown_label.setEnabled(True)
            self.dlg.uuid_pushButton.setEnabled(True)
            self.dlg.uuid_dropdown.setEnabled(True)
        else:
            self.dlg.uuid_label.setEnabled(False)
            self.dlg.uuid_manual_label.setEnabled(False)
            self.dlg.uuid_manual_entry.setEnabled(False)
            self.dlg.uuid_horizontalSlider.setEnabled(False)
            self.dlg.uuid_dropdown_label.setEnabled(False)
            self.dlg.uuid_pushButton.setEnabled(False)
            self.dlg.uuid_dropdown.setEnabled(False)


    def get_download_url(self,selected_collection,pop_uuid = False):
        #download_url = 'https://wzdev.crearecomputing.com/api/' + selected_collection + '?gis=geojson'
        download_url = 'https://api.weathercitizen.org/' + selected_collection + '?gis=geojson'
        
        # get times to import data
        start_time = self.dlg.dateTimeStart.dateTime()
        end_time = self.dlg.dateTimeEnd.dateTime()
        download_url = download_url + '&where={"properties.time":{"$gt":"' + start_time.toString('yyyy-MM-ddThh:mm:ss.zzzZ') + '","$lte":"' + end_time.toString('yyyy-MM-ddThh:mm:ss.zzzZ') + '"}'

        # select geometry
        loc_type = self.dlg.horizontalSlider.value()
        if loc_type == 0: # within square region
            lat_SW = self.dlg.lineEdit_lat_SW.text()
            lon_SW = self.dlg.lineEdit_lon_SW.text()
            lat_NE = self.dlg.lineEdit_lat_NE.text()
            lon_NE = self.dlg.lineEdit_lon_NE.text()
            coords = '[[' + lon_SW + ',' + lat_SW + '],[' + lon_NE + ',' + lat_NE + ']]'
            #sq_coords1 = '[' + lon_SW + ',' + lat_NE + ']'
            #sq_coords2 = '[' + lon_SW + ',' + lat_SW + ']'
            #sq_coords3 = '[' + lon_NE + ',' + lat_SW + ']'
            #sq_coords4 = '[' + lon_NE + ',' + lat_NE + ']'
            #geo_url = ',"geometry":{"$geoWithin":{"$geometry":{"type":"Polygon","coordinates":'
            #coords_url = '[[' + sq_coords1 + ',' + sq_coords2 + ',' + sq_coords3 + ',' + sq_coords4 + ',' + sq_coords1 + ']]'
            #crs_url = ',"crs":{"type":"name","properties":{"name":"urn:x-mongodb:crs:strictwinding:EPSG:4326"}}'
            #geo_url = geo_url + coords_url + crs_url + '}}}'
            geo_url = ',"geometry":{"$geoWithin":{"$box":' + coords + '}}'
            download_url = download_url + geo_url
        elif loc_type == 1: # distance from center point
            lat_center = self.dlg.lineEdit_lat_center.text()
            lon_center = self.dlg.lineEdit_lon_center.text()
            radius = self.dlg.lineEdit_radius.text()
            geo_url = ',"geometry":{"$near":{"$geometry":{"type":"Point","coordinates":['
            geo_url = geo_url + lon_center + ',' + lat_center + ']},"$maxDistance":' + radius + '}}'
            download_url = download_url + geo_url

        # select uuid - for geosensors only
        if selected_collection == "geosensors":
            uuid_type = self.dlg.uuid_horizontalSlider.value()
            if uuid_type == 0: # manual input
                selected_uuid = self.dlg.uuid_manual_entry.text()
            elif uuid_type == 1:
                print(uuids)
                selected_uuid = uuids[self.dlg.uuid_dropdown.currentIndex()]
            print('Getting data for uuid ', selected_uuid)
            if not (pop_uuid or selected_uuid == "All" or selected_uuid == "all" or selected_uuid == ""):
                uuid_url = ',"properties.uuid":"' + selected_uuid + '"'
                download_url = download_url + uuid_url


        # end URL
        download_url = download_url + '}'
        return download_url

    def populate_uuid_dropdown(self):
        global uuids 
        uuids = ['All']
        self.dlg.uuid_dropdown.clear()
        self.dlg.uuid_dropdown.addItems(uuids)
        self.num_wo_uuid = 0
        # get initial data
        selected_collection_idx = self.dlg.collectionBox.currentIndex()
        selected_collection = collections[selected_collection_idx]
        download_url = self.get_download_url(selected_collection, pop_uuid = True)
        r = requests.get(download_url)
        o = r.json()
        num_pages = int(math.ceil(o['_meta']['total']/200.))
        for i in range(len(o['features'])):
            if 'uuid' in o['features'][i]['properties']:
                if not o['features'][i]['properties']['uuid'] in uuids:
                    uuids.append(o['features'][i]['properties']['uuid'])
            else:
                self.num_wo_uuid = self.num_wo_uuid + 1
        for j in range(2, num_pages+1):
            download_url2 = download_url + '&page=' + str(j)
            r = requests.get(download_url2)
            o = r.json()
            for i in range(len(o['features'])):
                if 'uuid' in o['features'][i]['properties']:
                    if not o['features'][i]['properties']['uuid'] in uuids:
                        uuids.append(o['features'][i]['properties']['uuid'])
                else:
                    self.num_wo_uuid = self.num_wo_uuid+1
        self.dlg.uuid_dropdown.clear()
        self.dlg.uuid_dropdown.addItems(uuids)
        print(uuids)

    def update_numpages_label(self):
        self.dlg.label_numpages.setText('')
        # get collection type
        selected_collection_idx = self.dlg.collectionBox.currentIndex()
        selected_collection = collections[selected_collection_idx]
        download_url = self.get_download_url(selected_collection)
        print(download_url)
        # download data
        r = requests.get(download_url)
        o = r.json()
        num_pts = o['_meta']['total']
        num_pages = int(math.ceil(num_pts/200.))
        # display num_pages
        disp_text = str(num_pts) + ' points in ' + str(num_pages) + ' pages are available.'
        self.dlg.label_numpages.setText(disp_text)
        self.dlg.spinBox.setMaximum(num_pages)


    def run(self):
        """Run method that performs all the real work"""
        # setup the dialog
        if self.first_start == True:
            # load map if not already loaded
            self.load_map()
            # add collections to GUI
            self.dlg.header_label.setText('Import WeatherCitizen data from https://api.weathercitizen.org')
            self.dlg.collectionBox.clear()
            self.dlg.collectionBox.addItems(collections)
            # set default times based on current time - could update start time to be determined by start of collection time
            self.min_time = datetime(2017,8,31,20,0,0,tzinfo=pytz.UTC) # GUI displays 4 hours later for unknown reason
            self.max_time = datetime.utcnow()
            self.max_time = self.max_time.replace(tzinfo=pytz.UTC)
            self.dlg.dateTimeStart.setDateTimeRange(self.min_time, self.max_time)
            self.dlg.dateTimeEnd.setDateTimeRange(self.min_time, self.max_time)
            self.dlg.dateTimeStart.setDateTime(self.max_time - timedelta(days=7))
            self.dlg.dateTimeEnd.setDateTime(self.max_time)
            # set first_start flag to false
            self.first_start = False
        else:
            # update max time in subsequent setups - don't update display
            old_max_time = self.max_time
            self.max_time = datetime.utcnow()
            self.max_time = self.max_time.replace(tzinfo=pytz.UTC)
            self.dlg.dateTimeStart.setDateTimeRange(self.min_time, self.max_time)
            self.dlg.dateTimeEnd.setDateTimeRange(self.min_time, self.max_time)
            #print(self.dlg.dateTimeEnd.dateTime())
            #print(old_max_time)
            #if self.dlg.dateTimeEnd.dateTime() == old_max_time:
            #    self.dlg.dateTimeStart.setDateTime(self.max_time - timedelta(days=7))
            #    self.dlg.dateTimeEnd.setDateTime(self.max_time)
        # clear previous labels indicating number of pages available
        self.dlg.label_numpages.setText('')
        # update showing UUID input
        self.toggle_uuid_display()
        
        # show the dialog
        self.dlg.show()
        # live update showing the UUID input
        self.dlg.collectionBox.currentIndexChanged.connect(self.toggle_uuid_display)
        # live update populating the uuid dropdown list
        self.dlg.uuid_pushButton.clicked.connect(self.populate_uuid_dropdown)
        # live update the number of files to be downloaded
        self.dlg.pushButton.clicked.connect(self.update_numpages_label)
        
        # Run the dialog event loop
        result = self.dlg.exec_()
        # See if OK was pressed
        if result:
            # get inputs from GUI
            # get collection type and associated inputs
            selected_collection_idx = self.dlg.collectionBox.currentIndex()
            selected_collection = collections[selected_collection_idx]
            # get pages to import
            page = self.dlg.spinBox.value()
            checked = self.dlg.checkBox.isChecked()

            # get URL to download geodevice geojson data file
            download_url = self.get_download_url(selected_collection)
            if checked:
                page = 1
            if page > 1: # only single page download selected for a page > 1
                download_url = download_url + '&page=' + str(page)
            print(download_url)

            # download data
            r = requests.get(download_url)
            o = r.json()
            num_pages = int(math.ceil(o['_meta']['total']/200.))
            
            # reformat collections for flatten 'properties'
            f = o['features']
            for i in range(len(f)):
                # find nested properties
                nested_props = []
                list_props = []
                for p in f[i]['properties']: # scroll through properties
                    if type(f[i]['properties'][str(p)]) is dict and len(f[i]['properties'][str(p)]) > 0: # nested properties
                        nested_props.append(p)
                    elif type(f[i]['properties'][str(p)]) is list: # vector properties
                        list_props.append(p)
                # unpack lists (geosensors)
                for p in list_props:
                    for n in range(len(f[i]['properties'][str(p)])):
                        f[i]['properties'][unicode(p + '_' + str(n))] = f[i]['properties'][str(p)][n]
                # unpack dicts (openweather)
                for p in nested_props:
                    for p2 in f[i]['properties'][str(p)]: # scroll through nested properties
                        f[i]['properties'][unicode(p + '_' + p2)] = f[i]['properties'][str(p)][str(p2)]
                # remove nested versions
                for p in nested_props:
                    del f[i]['properties'][p]
                # remove list versions
                for p in list_props:
                    del f[i]['properties'][p]
                # add lat and lon to properties dict (future QGIS Attributes Table)
                if not('lat' in f[i]['properties']):
                    f[i]['properties'][unicode('lat')] = f[i]['geometry']['coordinates'][1]
                if not('lon' in f[i]['properties']):
                    f[i]['properties'][unicode('lon')] = f[i]['geometry']['coordinates'][0]
            
            # write json file and import to QGIS
            d = json.dumps(o)
            layer_label_prefix = selected_collection + '_'
            layer_label = layer_label_prefix + str(page) + 'of' + str(num_pages)
            self.iface.addVectorLayer(d, layer_label, 'ogr')

            # get coordinates to zoom map to new collection added
            layer = self.iface.activeLayer()
            pos2 = [math.radians(layer.extent().center()[0])*RADIUS, math.log(math.tan(math.pi/4 + math.radians(layer.extent().center()[1])/2))*RADIUS]
            offset_width = math.radians(layer.extent().width())*RADIUS
            offset_height = math.log(math.tan(math.pi/4 + math.radians(layer.extent().height())/2))*RADIUS
            west = pos2[0]-offset_width
            south = pos2[1]-offset_height
            east = pos2[0]+offset_width
            north = pos2[1]+offset_height
            num_pages = int(math.ceil(o['_meta']['total']/200.))

            # add subsequent layers
            if checked and num_pages > 1:
                for i in range(2,num_pages+1):
                    page = i
                    download_url2 = download_url + '&page=' + str(page)
                    print(download_url2)

                    # download data
                    r = requests.get(download_url2)
                    o = r.json()

                    # reformat collections for flatten 'properties'
                    f = o['features']
                    for i in range(len(f)):
                        # find nested properties
                        nested_props = []
                        list_props = []
                        for p in f[i]['properties']: # scroll through properties
                            if type(f[i]['properties'][str(p)]) is dict and len(f[i]['properties'][str(p)]) > 0: # nested properties
                                nested_props.append(p)
                            elif type(f[i]['properties'][str(p)]) is list: # vector properties
                                list_props.append(p)
                        # unpack lists (geosensors)
                        for p in list_props:
                            for n in range(len(f[i]['properties'][str(p)])):
                                f[i]['properties'][unicode(p + '_' + str(n))] = f[i]['properties'][str(p)][n]
                        # unpack dicts (openweather)
                        for p in nested_props:
                            for p2 in f[i]['properties'][str(p)]: # scroll through nested properties
                                f[i]['properties'][unicode(p + '_' + p2)] = f[i]['properties'][str(p)][str(p2)]
                        # remove nested versions
                        for p in nested_props:
                            del f[i]['properties'][p]
                        # remove list versions
                        for p in list_props:
                            del f[i]['properties'][p]
                        # add lat and lon to properties dict (future QGIS Attributes Table)
                        if not('lat' in f[i]['properties']):
                            f[i]['properties'][unicode('lat')] = f[i]['geometry']['coordinates'][1]
                        if not('lon' in f[i]['properties']):
                            f[i]['properties'][unicode('lon')] = f[i]['geometry']['coordinates'][0]
                    
                    # write json file and import to QGIS
                    d = json.dumps(o)
                    layer_label = layer_label_prefix + str(page) + 'of' + str(num_pages)
                    self.iface.addVectorLayer(d, layer_label, 'ogr')

                    # get new zoom coordinates
                    layer = self.iface.activeLayer()
                    pos2 = [math.radians(layer.extent().center()[0])*RADIUS, math.log(math.tan(math.pi/4 + math.radians(layer.extent().center()[1])/2))*RADIUS]
                    offset_width = math.radians(layer.extent().width())*RADIUS
                    offset_height = math.log(math.tan(math.pi/4 + math.radians(layer.extent().height())/2))*RADIUS
                    west = min(west, pos2[0]-offset_width)
                    south = min(south, pos2[1]-offset_height)
                    east = max(east, pos2[0]+offset_width)
                    north = max(north, pos2[1]+offset_height)

            #zoom to new data added
            zoomRectangle = QgsRectangle(west, south, east, north)
            canvas = self.iface.mapCanvas()
            canvas.setExtent(zoomRectangle)
            canvas.refresh()
